import numpy as np
import pytest
from numpy.random import PCG64, Generator

from bicfit import fit_damped_cosine, DampedCosineResult, NoOffset

testdata = [
    (1.0, 0.2, 0.05, np.pi / 2, 10.0, 150, 110),
    (2.3, 0.5, 0.1, 1.0, 3.3, 120, 90),
]


@pytest.mark.parametrize(
    "amplitude, pulsation, decay_rate, phase, offset, horizon, n_points", testdata
)
@pytest.mark.parametrize("post_fit", [False, True])
@pytest.mark.parametrize("noise,tol", [(0, 0.08), (0.05, 0.2)])
def test_single_damped_cosine(
    amplitude,
    pulsation,
    decay_rate,
    phase,
    offset,
    horizon,
    n_points,
    post_fit,
    noise,
    tol,
):
    rng = Generator(PCG64(42))
    times = np.linspace(0, horizon, n_points)
    noise_vec = rng.normal(0, noise, n_points) + 1j * rng.normal(0, noise, n_points)
    signal = (
        offset
        + amplitude * np.cos(phase + pulsation * times) * np.exp(-decay_rate * times)
        + noise_vec.real
    )

    result = fit_damped_cosine(times, signal, n_modes=1, post_fit=post_fit)
    try:
        assert abs(result.modes[0].amplitude - amplitude) / abs(amplitude) < tol, (
            "amplitude"
        )
        assert abs(result.modes[0].pulsation - pulsation) / pulsation < tol, "pulsation"
        assert abs(result.modes[0].decay_rate - decay_rate) / decay_rate < tol, (
            "decay_rate"
        )
        dphi = (result.modes[0].phase - phase + np.pi) % (2 * np.pi) - np.pi
        assert abs(dphi) < tol * np.pi, "phase"
        assert abs(result.offset - offset) / abs(offset) < tol, "offset"
    except AssertionError as e:
        failed_test = e.args[0].split("\n")[0]
        print(
            f"Failed estimating '{failed_test}' for \n"
            f"- amplitude   = {amplitude} (got {result.modes[0].amplitude})\n"
            f"- pulsation   = {pulsation} (got {result.modes[0].pulsation})\n"
            f"- decay_rate  = {decay_rate} (got {result.modes[0].decay_rate})\n"
            f"- phase       = {phase} (got {result.modes[0].phase})\n"
            f"- offset      = {offset} (got {result.offset})\n"
            f"- horizon     = {horizon}\n"
            f"- n_points    = {n_points}"
        )
        raise e


@pytest.mark.parametrize("post_fit", [False, True])
@pytest.mark.parametrize("noise,tol", [(0, 0.08), (0.05, 0.2)])
def test_two_damped_cosines(post_fit, noise, tol):
    rng = Generator(PCG64(42))
    n_points = 100
    times = np.linspace(0, 150, n_points)
    noise_vec = rng.normal(0, noise, n_points) + 1j * rng.normal(0, noise, n_points)

    offset = 5.2
    a1, a2 = 1.0, 0.6
    pulsation1, pulsation2 = 0.2, 0.4
    decay_rate1, decay_rate2 = 0.05, 0.01
    phase1, phase2 = 1.0, -0.2

    signal = offset
    signal += a1 * np.cos(phase1 + pulsation1 * times) * np.exp(-decay_rate1 * times)
    signal += a2 * np.cos(phase2 + pulsation2 * times) * np.exp(-decay_rate2 * times)
    signal += noise_vec.real

    result = fit_damped_cosine(times, signal, n_modes=2, post_fit=post_fit)
    try:
        m1, m2 = result.modes
        if m1.pulsation > m2.pulsation:
            m1, m2 = m2, m1
        assert abs(m1.amplitude - a1) / abs(a1) < tol, "a1"
        assert abs(m1.pulsation - pulsation1) / pulsation1 < tol, "pulsation1"
        assert abs(m1.decay_rate - decay_rate1) / decay_rate1 < tol, "decay_rate1"
        dphi1 = (m1.phase - phase1 + np.pi) % (2 * np.pi) - np.pi
        assert abs(dphi1) < tol * np.pi, "phase1"

        assert abs(m2.amplitude - a2) / abs(a2) < tol, "a2"
        assert abs(m2.pulsation - pulsation2) / pulsation2 < tol, "pulsation2"
        assert abs(m2.decay_rate - decay_rate2) / decay_rate2 < tol, "decay_rate2"
        dphi2 = (m2.phase - phase2 + np.pi) % (2 * np.pi) - np.pi
        assert abs(dphi2) < tol * np.pi, "phase2"

        assert abs(result.offset - offset) / abs(offset) < tol, "offset"
    except AssertionError as e:
        failed_test = e.args[0].split("\n")[0]
        print(
            f"Failed estimating '{failed_test}' for \n"
            f"- m1 = {m1}\n"
            f"- m2 = {m2}\n"
            f"- offset = {result.offset}"
        )
        raise e


def test_individual_mode():
    result = DampedCosineResult(
        offset=0.0,
        times=np.array([0.0, 1.0]),
        signal=np.array([2.0, 3.0]),
        amplitudes=np.array([4.0, 5.0, 6.0]),
        phases=np.array([7.0, 8.0, 9.0]),
        pulsations=np.array([10.0, 11.0, 12.0]),
        decay_rates=np.array([13.0, 14.0, 15.0]),
    )

    t = np.linspace(0, 10, 11)
    np.set_printoptions(linewidth=1000)
    assert result(t).shape == (11,)
    assert np.allclose(
        result(t), result.modes[0](t) + result.modes[1](t) + result.modes[2](t)
    )
    assert result(0.0).shape == tuple()
    assert np.isclose(
        result(0.0), result.modes[0](0.0) + result.modes[1](0.0) + result.modes[2](0.0)
    )


@pytest.mark.parametrize(
    "amplitude, pulsation, decay_rate, phase, _offset, horizon, n_points", testdata
)
def test_no_offset(amplitude, pulsation, decay_rate, phase, _offset, horizon, n_points):
    noise = 0.05
    tol = 0.2

    rng = Generator(PCG64(42))
    times = np.linspace(0, horizon, n_points)
    noise_vec = rng.normal(0, noise, n_points) + 1j * rng.normal(0, noise, n_points)
    signal = (
        amplitude * np.cos(phase + pulsation * times) * np.exp(-decay_rate * times)
        + noise_vec.real
    )

    result = fit_damped_cosine(times, signal, n_modes=1, post_fit=NoOffset())
    try:
        assert abs(result.modes[0].amplitude - amplitude) / abs(amplitude) < tol, (
            "amplitude"
        )
        assert abs(result.modes[0].pulsation - pulsation) / pulsation < tol, "pulsation"
        assert abs(result.modes[0].decay_rate - decay_rate) / decay_rate < tol, (
            "decay_rate"
        )
        dphi = (result.modes[0].phase - phase + np.pi) % (2 * np.pi) - np.pi
        assert abs(dphi) < tol * np.pi, "phase"
        assert abs(result.offset) == 0.0, "offset"
    except AssertionError as e:
        failed_test = e.args[0].split("\n")[0]
        print(
            f"Failed estimating '{failed_test}' for \n"
            f"- amplitude   = {amplitude} (got {result.modes[0].amplitude})\n"
            f"- pulsation   = {pulsation} (got {result.modes[0].pulsation})\n"
            f"- decay_rate  = {decay_rate} (got {result.modes[0].decay_rate})\n"
            f"- phase       = {phase} (got {result.modes[0].phase})\n"
            f"- offset      = 0.0 (got {result.offset})\n"
            f"- horizon     = {horizon}\n"
            f"- n_points    = {n_points}"
        )
        raise e



import numpy as np
import pytest
from numpy.random import Generator, PCG64

from bicfit import fit_exponential_decay, ExponentialDecayResult, NoOffset

real_testdata = [
    (1.0, 0.05, 10.0, 150, 110),
    (-2.3, 0.1, 3.3, 120, 90),
]

complex_testdata = [
    (1.0 + 0.2j, 0.05, 10.0 + 5j, 150, 110),
    (2.3 - 1.4j, 0.1, 3.3 - 2j, 120, 90),
    (-4.3 - 1.0j, 0.15, 3.3 - 2j, 100, 90),
]


@pytest.mark.parametrize(
    "amplitude, decay_rate, offset, horizon, n_points", real_testdata
)
@pytest.mark.parametrize(
    "post_fit,noise,tol",
    [(False, 0, 0.02), (False, 0.05, 0.2), (True, 0, 0.02), (True, 0.05, 0.1)],
)
def test_single_real_exponential(
    amplitude, decay_rate, offset, horizon, n_points, post_fit, noise, tol
):
    rng = Generator(PCG64(42))
    times = np.linspace(0, horizon, n_points)
    noise_vec = rng.normal(0, noise, n_points)
    signal = offset + amplitude * np.exp(-decay_rate * times) + noise_vec

    result = fit_exponential_decay(
        times, signal, n_modes=1, post_fit=post_fit, is_complex=False
    )
    try:
        assert abs(result.modes[0].amplitude - amplitude) / abs(amplitude) < tol, (
            "amplitude"
        )
        assert abs(result.modes[0].decay_rate - decay_rate) / decay_rate < tol, (
            "decay_rate"
        )
        assert abs(result.offset - offset) / abs(offset) < tol, "offset"
    except AssertionError as e:
        failed = e.args[0]
        print(
            f"Failed estimating '{failed}' for real exponential:\n"
            f"- amplitude true  = {amplitude} got={result.modes[0].amplitude}\n"
            f"- decay_rate true = {decay_rate} got={result.modes[0].decay_rate}\n"
            f"- offset true     = {offset} got={result.offset}\n"
            f"- horizon         = {horizon}\n"
            f"- n_points        = {n_points}"
        )
        raise e


@pytest.mark.parametrize(
    "amplitude, decay_rate, offset, horizon, n_points", complex_testdata
)
@pytest.mark.parametrize(
    "post_fit, noise,tol",
    [(False, 0, 0.03), (False, 0.05, 0.2), (True, 0, 0.03), (True, 0.05, 0.1)],
)
def test_single_complex_exponential(
    amplitude, decay_rate, offset, horizon, n_points, post_fit, noise, tol
):
    rng = Generator(PCG64(42))
    times = np.linspace(0, horizon, n_points)
    noise_vec = rng.normal(0, noise, n_points) + 1j * rng.normal(0, noise, n_points)
    signal = offset + amplitude * np.exp(-decay_rate * times) + noise_vec

    result = fit_exponential_decay(
        times, signal, n_modes=1, post_fit=post_fit, is_complex=True
    )
    amp_est = result.modes[0].amplitude

    try:
        assert abs(abs(amp_est) - abs(amplitude)) / abs(amplitude) < tol, "amplitude"
        assert abs(result.modes[0].decay_rate - decay_rate) / decay_rate < tol, (
            "decay_rate"
        )
        assert abs(result.offset - offset) / abs(offset) < tol, "offset"
    except AssertionError as e:
        failed = e.args[0]
        print(
            f"Failed estimating '{failed}' for complex exponential:\n"
            f"- amplitude true={amplitude} got={amp_est}\n"
            f"- decay_rate     true={decay_rate} got={result.modes[0].decay_rate}\n"
            f"- offset    true={offset} got={result.offset}\n"
            f"- horizon   = {horizon}\n"
            f"- n_points  = {n_points}"
        )
        raise e


@pytest.mark.parametrize(
    "post_fit,noise,tol",
    [(False, 0, 0.05), (False, 0.01, 0.6), (True, 0, 0.01), (True, 0.01, 0.1)],
)
def test_two_real_exponentials(post_fit, noise, tol):
    rng = Generator(PCG64(42))
    n_points = 120
    times = np.linspace(0, 150, n_points)
    noise_vec = rng.normal(0, noise, n_points)

    offset = 4.5
    a1, a2 = 0.6, 1.0
    decay_rate1, decay_rate2 = 0.01, 0.1

    signal = (
        offset
        + a1 * np.exp(-decay_rate1 * times)
        + a2 * np.exp(-decay_rate2 * times)
        + noise_vec
    )

    result = fit_exponential_decay(
        times, signal, n_modes=2, post_fit=post_fit, is_complex=False
    )

    try:
        assert abs(result.modes[0].amplitude - a1) / abs(a1) < tol, "a1"
        assert abs(result.modes[0].decay_rate - decay_rate1) / decay_rate1 < tol, (
            "decay_rate1"
        )

        assert abs(result.modes[1].amplitude - a2) / abs(a2) < tol, "a2"
        assert abs(result.modes[1].decay_rate - decay_rate2) / decay_rate2 < tol, (
            "decay_rate2"
        )

        assert abs(result.offset - offset) / abs(offset) < tol, "offset"
    except AssertionError as e:
        failed_test = e.args[0].split("\n")[0]
        print(
            f"Failed estimating '{failed_test}' for \n"
            f"- amplitude 1  = {result.modes[0].amplitude}\n"
            f"- decay_rate 1 = {result.modes[0].decay_rate})\n"
            f"- amplitude 2  = {result.modes[1].amplitude}\n"
            f"- decay_rate 2 = {result.modes[1].decay_rate})\n"
            f"- offset       = {result.offset})\n"
        )
        raise e


@pytest.mark.parametrize(
    "post_fit,noise,tol",
    [(False, 0, 0.05), (False, 0.01, 0.3), (True, 0, 0.01), (True, 0.01, 0.05)],
)
def test_two_complex_exponentials(post_fit, noise, tol):
    rng = Generator(PCG64(42))
    n_points = 120
    times = np.linspace(0, 150, n_points)
    noise_vec = rng.normal(0, noise, n_points) + 1j * rng.normal(0, noise, n_points)

    offset = 4.5 + 2.2j
    a1, a2 = 1.0 + 0.2j, 0.6 - 0.1j
    decay_rate1, decay_rate2 = 0.02, 0.1

    signal = (
        offset
        + a1 * np.exp(-decay_rate1 * times)
        + a2 * np.exp(-decay_rate2 * times)
        + noise_vec
    )

    result = fit_exponential_decay(
        times, signal, n_modes=2, post_fit=post_fit, is_complex=True
    )
    try:
        assert abs(result.modes[0].amplitude - a1) / abs(a1) < tol, "a1"
        assert abs(result.modes[0].decay_rate - decay_rate1) / decay_rate1 < tol, (
            "decay_rate1"
        )

        assert abs(result.modes[1].amplitude - a2) / abs(a2) < tol, "a2"
        assert abs(result.modes[1].decay_rate - decay_rate2) / decay_rate2 < tol, (
            "decay_rate2"
        )

        assert abs(result.offset - offset) / abs(offset) < tol, "offset"
    except AssertionError as e:
        failed_test = e.args[0].split("\n")[0]
        print(
            f"Failed estimating '{failed_test}' for \n"
            f"- amplitude 1  = {result.modes[0].amplitude}\n"
            f"- decay_rate 1 = {result.modes[0].decay_rate})\n"
            f"- amplitude 2  = {result.modes[1].amplitude}\n"
            f"- decay_rate 2 = {result.modes[1].decay_rate})\n"
            f"- offset       = {result.offset})\n"
        )
        raise e


def test_individual_mode():
    result = ExponentialDecayResult(
        0.0,
        np.array([0.0, 1.0]),
        np.array([1.0, 2.0]),
        np.array([3.0, 4.0, 5.0]),
        np.array([6.0, 7.0, 8.0]),
    )

    t = np.linspace(0, 10, 11)
    assert result(t).shape == (11,)
    assert np.allclose(
        result(t), result.modes[0](t) + result.modes[1](t) + result.modes[2](t)
    )
    assert result(0.0).shape == tuple()
    assert np.isclose(
        result(0.0), result.modes[0](0.0) + result.modes[1](0.0) + result.modes[2](0.0)
    )


@pytest.mark.parametrize(
    "amplitude, decay_rate, offset, horizon, n_points", complex_testdata
)
@pytest.mark.parametrize("noise,tol", [(0, 0.03), (0.05, 0.1)])
def test_no_offset(amplitude, decay_rate, offset, horizon, n_points, noise, tol):
    rng = Generator(PCG64(42))
    times = np.linspace(0, horizon, n_points)
    noise_vec = rng.normal(0, noise, n_points) + 1j * rng.normal(0, noise, n_points)
    signal = amplitude * np.exp(-decay_rate * times) + noise_vec

    result = fit_exponential_decay(
        times, signal, n_modes=1, post_fit=NoOffset(), is_complex=True
    )
    amp_est = result.modes[0].amplitude

    try:
        assert abs(abs(amp_est) - abs(amplitude)) / abs(amplitude) < tol, "amplitude"
        assert abs(result.modes[0].decay_rate - decay_rate) / decay_rate < tol, (
            "decay_rate"
        )
        assert abs(result.offset) < 1e-5, "offset"
    except AssertionError as e:
        failed = e.args[0]
        print(
            f"Failed estimating '{failed}' for complex exponential:\n"
            f"- amplitude true={amplitude} got={amp_est}\n"
            f"- decay_rate     true={decay_rate} got={result.modes[0].decay_rate}\n"
            f"- offset    true={offset} got={result.offset}\n"
            f"- horizon   = {horizon}\n"
            f"- n_points  = {n_points}"
        )
        raise e



import functools as ft
from dataclasses import dataclass
from typing import Callable, Tuple

import numpy as np
from scipy.optimize import minimize

from .models import _damped_cosine_model, _complex_exponential_model, _exponential_model
from .results import DampedCosineResult
from .types import FloatLike

# =====================================================================================================================
# Post fit options
# =====================================================================================================================


@dataclass
class NoOffset:
    pass


_PostFitOptions = None | NoOffset

# =====================================================================================================================
# Common functions
# =====================================================================================================================


def _cost(
    parameters: np.ndarray,
    times: np.ndarray,
    signal: np.ndarray,
    model: Callable[[np.ndarray, np.ndarray], np.ndarray],
) -> float:
    model_signal = model(times, parameters)
    return np.sum(np.abs(model_signal - signal) ** 2)


# =====================================================================================================================
# Complex exponential model
# =====================================================================================================================


def _complex_exponential_adapter(
    t: np.ndarray[float], x: np.ndarray[FloatLike], is_there_offset: bool = True
):
    if is_there_offset:
        offset = x[0] + 1j * x[1]
        x = x[2:]
    else:
        offset = complex(0.0)
    modes = x.reshape(-1, 4)

    amplitudes_re, amplitudes_im, pulsations, decay_rates = (
        modes[:, 0],
        modes[:, 1],
        modes[:, 2],
        modes[:, 3],
    )
    decay_rates = np.abs(decay_rates)  # enforce the positivity of the decay rate
    amplitudes = amplitudes_re + 1j * amplitudes_im
    return _complex_exponential_model(t, offset, amplitudes, pulsations, decay_rates)


def _post_fit_complex_exponential(
    times: np.ndarray,
    signal: np.ndarray,
    offset: complex,
    amplitudes: np.ndarray[complex],
    pulsations: np.ndarray[float],
    decay_rates: np.ndarray[float],
    options: _PostFitOptions,
) -> Tuple[complex, np.ndarray[complex], np.ndarray[float], np.ndarray[float]]:
    assert len(amplitudes) == len(pulsations) == len(decay_rates)
    is_there_offset = not isinstance(options, NoOffset)

    cost = ft.partial(
        _cost,
        times=times,
        signal=signal,
        model=ft.partial(_complex_exponential_adapter, is_there_offset=is_there_offset),
    )

    x0_offset = []
    if is_there_offset:
        x0_offset = [offset.real, offset.imag]
    x0_modes = np.stack(
        (amplitudes.real, amplitudes.imag, pulsations, decay_rates)
    ).T.flatten()
    x0 = np.concatenate((x0_offset, x0_modes))

    xopt = minimize(cost, x0).x

    if is_there_offset:
        offset = xopt[0] + 1j * xopt[1]
        xopt = xopt[2:]
    else:
        offset = complex(0.0)
    amplitudes_re, amplitudes_im, pulsations, decay_rates = xopt.reshape(-1, 4).T
    amplitudes = amplitudes_re + 1j * amplitudes_im
    decay_rates = np.abs(decay_rates)  # Enforce the positivity of the decay rates

    return offset, amplitudes, pulsations, decay_rates


# =====================================================================================================================
# Real exponential model
# =====================================================================================================================


def _exponential_adapter(
    t: np.ndarray[float],
    x: np.ndarray[FloatLike],
    is_complex: bool,
    is_there_offset: bool = True,
) -> np.ndarray[FloatLike]:
    if is_complex:
        if is_there_offset:
            offset = complex(x[0] + 1j * x[1])
            x = x[2:]
        else:
            offset = complex(0.0)
        modes = x.reshape(-1, 3)
        amplitudes_re, amplitudes_im, decay_rates = (
            modes[:, 0],
            modes[:, 1],
            modes[:, 2],
        )
        amplitudes = amplitudes_re + 1j * amplitudes_im
    else:
        if is_there_offset:
            offset = complex(x[0])
            x = x[1:]
        else:
            offset = complex(0.0)
        modes = x.reshape(-1, 2)
        amplitudes, decay_rates = modes[:, 0], modes[:, 1]

    decay_rates = np.abs(decay_rates)  # Enforce the positivity of the decay rate
    return _exponential_model(t, offset, amplitudes, decay_rates)


def _post_fit_exponential(
    times: np.ndarray,
    signal: np.ndarray,
    offset: complex,
    amplitudes: np.ndarray[complex],
    decay_rates: np.ndarray[complex],
    is_complex: bool,
    options: _PostFitOptions,
) -> Tuple[FloatLike, np.ndarray[FloatLike], np.ndarray[FloatLike]]:
    assert len(amplitudes) == len(decay_rates)
    is_there_offset = not isinstance(options, NoOffset)

    cost = ft.partial(
        _cost,
        times=times,
        signal=signal,
        model=ft.partial(
            _exponential_adapter, is_complex=is_complex, is_there_offset=is_there_offset
        ),
    )

    x0_offset = []
    if is_there_offset:
        if is_complex:
            x0_offset = [offset.real, offset.imag]
        else:
            x0_offset = [offset.real]
    if is_complex:
        x0_modes = np.stack((amplitudes.real, amplitudes.imag, decay_rates)).T.flatten()
    else:
        x0_modes = np.stack((amplitudes.real, decay_rates)).T.flatten()
    x0 = np.concatenate((x0_offset, x0_modes))

    xopt = minimize(cost, x0).x

    offset = complex(0.0)
    if is_there_offset:
        if is_complex:
            offset = xopt[0] + 1j * xopt[1]
            xopt = xopt[2:]
        else:
            offset = xopt[0]
            xopt = xopt[1:]
    if is_complex:
        amplitudes_re, amplitudes_im, decay_rates = xopt.reshape(-1, 3).T
        amplitudes = amplitudes_re + 1j * amplitudes_im
    else:
        amplitudes, decay_rates = xopt.reshape(-1, 2).T
    decay_rates = np.abs(decay_rates)  # Enforce the positivity of the decay rates

    return offset, amplitudes, decay_rates


# =====================================================================================================================
# Damped cosine model
# =====================================================================================================================


def _damped_cosine_adapter(
    t: np.ndarray[float], x: np.ndarray[float], is_there_offset: bool = True
) -> np.ndarray[float]:
    if is_there_offset:
        offset = x[0]
        x = x[1:]
    else:
        offset = 0.0
    modes = x.reshape(-1, 4)
    amplitudes, phases, pulsations, decay_rates = (
        modes[:, 0],
        modes[:, 1],
        modes[:, 2],
        modes[:, 3],
    )

    decay_rates = np.abs(decay_rates)  # Enforce the positivity of the decay rate
    return _damped_cosine_model(t, offset, amplitudes, phases, pulsations, decay_rates)


def _post_fit_damped_cosine(
    times: np.ndarray[float],
    signal: np.ndarray[float],
    offset: complex,
    amplitudes: np.ndarray[float],
    phases: np.ndarray[float],
    pulsations: np.ndarray[float],
    decay_rates: np.ndarray[float],
    options: _PostFitOptions,
) -> DampedCosineResult:
    is_there_offset = not isinstance(options, NoOffset)

    cost = ft.partial(
        _cost,
        times=times,
        signal=signal,
        model=ft.partial(_damped_cosine_adapter, is_there_offset=is_there_offset),
    )

    x0_offset = []
    if is_there_offset:
        x0_offset = [offset]
    x0_modes = np.stack((amplitudes, phases, pulsations, decay_rates)).T.flatten()
    x0 = np.concatenate((x0_offset, x0_modes))

    xopt = minimize(cost, x0).x

    offset = 0.0
    if is_there_offset:
        offset = xopt[0]
        xopt = xopt[1:]
    amplitudes, phases, pulsations, decay_rates = xopt.reshape(-1, 4).T
    decay_rates = np.abs(decay_rates)  # Enforce the positivity of the decay rate

    new_result = DampedCosineResult(
        times=times,
        signal=signal,
        offset=offset,
        amplitudes=amplitudes,
        pulsations=pulsations,
        decay_rates=decay_rates,
        phases=phases,
    )

    return new_result



from numpy.random import Generator, PCG64
import numpy as np
import pytest

from bicfit import fit_complex_exponential, ComplexResult, NoOffset

testdata = [
    (1.0, 13.7, 0.92, 61 + 62j, 3.6, 110),
    (21, 2, 0.01, 1, 50, 100),
]


@pytest.mark.parametrize(
    "amplitude, pulsation, decay_rate, offset, horizon, n_points", testdata
)
@pytest.mark.parametrize("post_fit", [False, True])
@pytest.mark.parametrize("noise,tol", [(0, 0.01), (0.1, 0.15)])
def test_single_exponential(
    amplitude, decay_rate, pulsation, offset, horizon, n_points, post_fit, noise, tol
):
    rng = Generator(PCG64(42))
    times = np.linspace(0, horizon, n_points)
    noise = rng.normal(0, noise, n_points) + 1j * rng.normal(0, noise, n_points)
    signal = offset + amplitude * np.exp((1j * pulsation - decay_rate) * times) + noise

    # fit the signal
    result = fit_complex_exponential(times, signal, n_modes=1, post_fit=post_fit)
    try:
        assert abs(result.modes[0].amplitude - amplitude) / abs(amplitude) < tol, (
            "amplitude"
        )
        assert abs(result.modes[0].pulsation - pulsation) / pulsation < tol, "pulsation"
        assert abs(result.modes[0].decay_rate - decay_rate) / decay_rate < tol, (
            "decay_rate"
        )
        assert abs(result.offset - offset) / abs(offset) < tol, "offset"
    except AssertionError as e:
        failed_test = e.args[0].split("\n")[0]
        print(
            f"Failed estimating '{failed_test}' for \n"
            f"- amplitude  = {amplitude} (got {result.modes[0].amplitude})\n"
            f"- pulsation  = {pulsation} (got {result.modes[0].pulsation})\n"
            f"- decay_rate = {decay_rate} (got {result.modes[0].decay_rate})\n"
            f"- offset     = {offset} (got {result.offset})\n"
            f"- horizon    = {horizon}\n"
            f"- n_points   = {n_points}"
        )
        raise e


@pytest.mark.parametrize(
    "post_fit,noise,tol",
    [(False, 0, 0.01), (False, 0.1, 0.25), (True, 0, 0.01), (True, 0.1, 0.2)],
)
def test_two_exponential(post_fit, noise, tol):
    rng = Generator(PCG64(42))
    n_points = 100
    times = np.linspace(0, 150, n_points)
    noise = rng.normal(0, noise, n_points) + 1j * rng.normal(0, noise, n_points)

    offset = 10.2 + 20.3j
    a1, a2 = 0.5, 1.0
    pulsation1, pulsation2 = 0.4, 0.2
    decay_rate1, decay_rate2 = 0.02, 0.05

    signal = offset
    signal += a1 * np.exp((1j * pulsation1 - decay_rate1) * times)
    signal += a2 * np.exp((1j * pulsation2 - decay_rate2) * times)
    signal += noise

    # fit the signal
    result = fit_complex_exponential(times, signal, n_modes=2, post_fit=post_fit)

    try:
        assert abs(result.modes[0].amplitude - a1) / a1 < tol, "a1"
        assert abs(result.modes[0].pulsation - pulsation1) / pulsation1 < tol, (
            "pulsation1"
        )
        assert abs(result.modes[0].decay_rate - decay_rate1) / decay_rate1 < tol, (
            "decay_rate1"
        )

        assert abs(result.modes[1].amplitude - a2) / a2 < tol, "a2"
        assert abs(result.modes[1].pulsation - pulsation2) / pulsation2 < tol, (
            "pulsation2"
        )
        assert abs(result.modes[1].decay_rate - decay_rate2) / decay_rate2 < tol, (
            "decay_rate2"
        )

        assert abs(result.offset - offset) / abs(offset) < tol, "offset"
    except AssertionError as e:
        failed_test = e.args[0].split("\n")[0]
        print(
            f"Failed estimating '{failed_test}' for \n"
            f"- amplitude 1   = {result.modes[0].amplitude}\n"
            f"- pulsation 1   = {result.modes[0].pulsation}\n"
            f"- decay_rate 1  = {result.modes[0].decay_rate})\n"
            f"- amplitude 2   = {result.modes[1].amplitude}\n"
            f"- pulsation 2   = {result.modes[1].pulsation}\n"
            f"- decay_rate 2  = {result.modes[1].decay_rate})\n"
            f"- offset        = {result.offset})\n"
        )
        raise e


def test_individual_mode():
    result = ComplexResult(
        0.0,
        np.array([0.0, 1.0]),
        np.array([2.0, 3.0]),
        np.array([4.0, 5.0, 6.0]),
        np.array([7.0, 8.0, 9.0]),
        np.array([10.0, 11.0, 12.0]),
    )

    t = np.linspace(0, 10, 11)
    assert result(t).shape == (11,)
    assert np.allclose(
        result(t), result.modes[0](t) + result.modes[1](t) + result.modes[2](t)
    )
    assert result(0.0).shape == tuple()
    assert np.isclose(
        result(0.0), result.modes[0](0.0) + result.modes[1](0.0) + result.modes[2](0.0)
    )


@pytest.mark.parametrize(
    "amplitude, pulsation, decay_rate, _offset, horizon, n_points", testdata
)
def test_no_offset(amplitude, decay_rate, pulsation, _offset, horizon, n_points):
    rng = Generator(PCG64(42))
    noise, tol = 0.1, 0.15
    times = np.linspace(0, horizon, n_points)
    noise = rng.normal(0, noise, n_points) + 1j * rng.normal(0, noise, n_points)
    signal = amplitude * np.exp((1j * pulsation - decay_rate) * times) + noise

    # fit the signal
    result = fit_complex_exponential(times, signal, n_modes=1, post_fit=NoOffset())
    try:
        assert abs(result.modes[0].amplitude - amplitude) / abs(amplitude) < tol, (
            "amplitude"
        )
        assert abs(result.modes[0].pulsation - pulsation) / pulsation < tol, "pulsation"
        assert abs(result.modes[0].decay_rate - decay_rate) / decay_rate < tol, (
            "decay_rate"
        )
        assert abs(result.offset) < 1e-5, "offset"
    except AssertionError as e:
        failed_test = e.args[0].split("\n")[0]
        print(
            f"Failed estimating '{failed_test}' for \n"
            f"- amplitude  = {amplitude} (got {result.modes[0].amplitude})\n"
            f"- pulsation  = {pulsation} (got {result.modes[0].pulsation})\n"
            f"- decay_rate = {decay_rate} (got {result.modes[0].decay_rate})\n"
            f"- offset     = 0.0 (got {result.offset})\n"
            f"- horizon    = {horizon}\n"
            f"- n_points   = {n_points}"
        )
        raise e



import numpy as np
import pytest

from bicfit import _match_real_modes


def test_match_real_modes_happy_path():
    amplitudes = np.array([1 + 2j, 1 - 2j, 3 + 4j, 3 - 4j, 5 + 6j, 5 - 6j])
    pulsations = np.array([10.0, -10.0, 20.0, -20.0, 30.0, -30.0])
    decay_rates = np.array([5.0, 5.0, 7.0, 7.0, 9.0, 9.0])
    tol = 1e-9

    real_amplitudes, real_phases, real_pulsations, real_decay_rates = _match_real_modes(
        amplitudes, pulsations, decay_rates, tol
    )

    assert real_pulsations.shape == (3,)
    assert real_decay_rates.shape == (3,)
    assert real_phases.shape == (3,)

    assert np.allclose(real_pulsations, [10.0, 20.0, 30.0])
    assert np.allclose(real_decay_rates, [5.0, 7.0, 9.0])

    expected_phases = [np.angle(1 + 2j), np.angle(3 + 4j), np.angle(5 + 6j)]
    assert np.allclose(real_phases, expected_phases)

    expected_real_amplitude = np.array(
        [2 * abs(1 + 2j), 2 * abs(3 + 4j), 2 * abs(5 + 6j)]
    )
    assert np.allclose(real_amplitudes, expected_real_amplitude)


def test_match_real_modes_frequency_mismatch():
    amplitudes = np.array([1 + 2j, 3 + 4j, 1 - 2j, 3 - 4j])

    pulsations = np.array([10.0, 20.0, -10.01, -20.0])
    decay_rates = np.array([5.0, 7.0, 5.0, 7.0])
    tol = 1e-3
    with pytest.raises(RuntimeError):
        _match_real_modes(amplitudes, pulsations, decay_rates, tol)


def test_match_real_modes_decay_rate_mismatch():
    amplitudes = np.array([1 + 2j, 3 + 4j, 1 - 2j, 3 - 4j])
    pulsations = np.array([10.0, 20.0, -10.0, -20.0])
    decay_rates = np.array([5.0, 7.0, 5.002, 7.0])
    tol = 1e-3
    with pytest.raises(RuntimeError):
        _match_real_modes(amplitudes, pulsations, decay_rates, tol)


def test_match_real_modes_amplitude_mismatch():
    amplitudes = np.array([1 + 2j, 3 + 4j, 1 - 2.01j, 3 - 4j])
    pulsations = np.array([10.0, 20.0, -10.0, -20.0])
    decay_rates = np.array([5.0, 7.0, 5.0, 7.0])
    tol = 5e-3
    with pytest.raises(RuntimeError):
        _match_real_modes(amplitudes, pulsations, decay_rates, tol)
